/*$Id: mgtype.c,v 1.26 2000/09/25 19:56:03 curfman Exp $*/

#include "petscsys.h"
#include "petscmat.h"       /*I "petscmat.h"  I*/

#undef __FUNC__  
#define __FUNC__ /*<a name=""></a>*/"MatGetTypeFromOptions"
/*@C
   MatGetTypeFromOptions - Determines from the options database what matrix
   format the user has specified.

   Collective over MPI_Comm

   Input Parameter:
+  comm - the MPI communicator
.  type - the type of matrix desired, for example MATSEQAIJ, MATMPIAIJ
-  pre - optional string to prepend to the name

   Output Parameters:
.  set - flag indicating whether user set matrix type option; or PETSC_NULL.

   Basic Options Database Keys:
   These options return MATSEQxxx or MATMPIxxx,
   depending on the communicator, comm.
+    -mat_aij      - AIJ type
.    -mat_baij     - block AIJ type
.    -mat_sbaij    - symmetric block AIJ type
.    -mat_dense    - dense type
-    -mat_bdiag    - block diagonal type

   More Options Database Keys:
+    -mat_seqaij   - MATSEQAIJ
.    -mat_mpiaij   - MATMPIAIJ
.    -mat_seqbaij  - MATSEQBAIJ
.    -mat_mpibaij  - MATMPIBAIJ
.    -mat_seqsbaij - MATSEQSBAIJ
.    -mat_mpisbaij - MATMPISBAIJ
.    -mat_seqbdiag - MATSEQBDAIG
.    -mat_mpibdiag - MATMPIBDIAG
.    -mat_mpirowbs - MATMPIROWBS
.    -mat_seqdense - MATSEQDENSE
-    -mat_mpidense - MATMPIDENSE

   Notes:
   This routine is automatically called within MatCreate() and MatLoad().

   Level: developer

.keywords: matrix, get, format, from, options

.seealso: MatCreate(), MatLoad()
@*/

int MatGetTypeFromOptions(MPI_Comm comm,char *pre,MatType *type,PetscTruth *set)
{
  int               size,ierr;
  PetscTruth        flg1,flg2,flg3,flg4,flg5,flg6,flg7,flg8,flg9,flg10;
  PetscTruth        flg11,flg12,flg13,flg14,flg15,flg16,flg17;
  static PetscTruth helpprinted = PETSC_FALSE;
  char              p[64];

  PetscFunctionBegin;
  PetscValidIntPointer(type); if (set) PetscValidIntPointer(set);

  ierr = PetscStrcpy(p,"-");CHKERRQ(ierr);
  if (pre) {ierr = PetscStrcat(p,pre);CHKERRQ(ierr);}

  ierr = MPI_Comm_size(comm,&size);CHKERRQ(ierr);
  ierr = OptionsHasName(PETSC_NULL,"-help",&flg1);CHKERRQ(ierr);
  if (flg1 && !helpprinted) {
    ierr = (*PetscHelpPrintf)(comm,"Matrix format options:\n");CHKERRQ(ierr);
    ierr = (*PetscHelpPrintf)(comm,"  %smat_aij, %smat_seqaij, %smat_mpiaij\n",p,p,p);CHKERRQ(ierr);
    ierr = (*PetscHelpPrintf)(comm,"  %smat_baij, %smat_seqbaij, %smat_mpibaij\n",p,p,p);CHKERRQ(ierr);
    ierr = (*PetscHelpPrintf)(comm,"  %smat_sbaij, %smat_seqsbaij, %smat_mpisbaij\n",p,p,p);CHKERRQ(ierr);
    ierr = (*PetscHelpPrintf)(comm,"  %smat_dense, %smat_seqdense, %smat_mpidense\n",p,p,p);CHKERRQ(ierr);
    ierr = (*PetscHelpPrintf)(comm,"  %smat_mpirowbs, %smat_bdiag, %smat_seqbdiag, %smat_mpibdiag\n",p,p,p,p);CHKERRQ(ierr);
    helpprinted = PETSC_TRUE;
  }
  ierr = OptionsHasName(pre,"-mat_seqdense",&flg1);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_mpidense",&flg2);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_dense",&flg3);CHKERRQ(ierr);

  ierr = OptionsHasName(pre,"-mat_seqbdiag",&flg4);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_mpibdiag",&flg5);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_bdiag",&flg6);CHKERRQ(ierr);

  ierr = OptionsHasName(pre,"-mat_seqbaij",&flg7);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_mpibaij",&flg8);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_baij",&flg9);CHKERRQ(ierr);

  ierr = OptionsHasName(pre,"-mat_seqaij",&flg10);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_mpiaij",&flg11);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_aij",&flg12);CHKERRQ(ierr);

  ierr = OptionsHasName(pre,"-mat_seqsbaij",&flg13);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_mpisbaij",&flg14);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_sbaij",&flg15);CHKERRQ(ierr);

  ierr = OptionsHasName(pre,"-mat_mpirowbs",&flg16);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_csr",&flg17);CHKERRQ(ierr);

  if (set) *set = PETSC_TRUE;

  if (flg1) {
    *type = MATSEQDENSE;
  }  else if (flg2) {
    *type = MATMPIDENSE;
  }  else if (flg3) {
    if (size == 1) *type = MATSEQDENSE;
    else *type = MATMPIDENSE;
  } else if (flg4) {
    *type = MATSEQBDIAG;
  } else if (flg5) {
    *type = MATMPIBDIAG;
  } else if (flg6) {
    if (size == 1) *type = MATSEQBDIAG;
    else *type = MATMPIBDIAG;
  } else if (flg7) {
    *type = MATSEQBAIJ;
  } else if (flg8) {
    *type = MATMPIBAIJ;
  } else if (flg9) {
    if (size == 1) *type = MATSEQBAIJ;
    else *type = MATMPIBAIJ;
  } else if (flg10) {
    *type = MATSEQAIJ;
  } else if (flg11) {
    *type = MATMPIAIJ;
  } else if (flg12) {
    if (size == 1) *type = MATSEQAIJ;
    else *type = MATMPIAIJ;
  } else if (flg13) {
    *type = MATSEQSBAIJ;
  } else if (flg14) {
    *type = MATMPISBAIJ;
  } else if (flg15) {
    if (size == 1) *type = MATSEQSBAIJ;
    else *type = MATMPISBAIJ;
  } else if (flg16) {
    *type = MATMPIROWBS;
  } else if (flg17) {
    *type = MATMPIADJ;
  } else {
    if (size == 1) *type = MATSEQAIJ;
    else *type = MATMPIAIJ;
    if (set) *set = PETSC_FALSE;
  }
  PetscFunctionReturn(0);
}
